/*
** Module   : OutputExtentParameter.java
** Abstract : Special wrapper used for variables passed as extent parameters to functions or
**            procedures. 
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20130209 Created initial version.
*/

package com.goldencode.p2j.util;

import java.util.*;
import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.security.*;

/**
 * Special wrapper used for variables passed as extent parameters to functions or procedures.
 * On normal function and procedure exit, if the reference for an output or input-output extent
 * parameter has changed, this will update the caller's variable reference too.
 */
public abstract class OutputExtentParameter<T extends BaseDataType>
{
   /** Object which manages all instances of this class by context */
   private static final ContextLocal<Context> context =
   new ContextLocal<Context>()
   {
      protected Context initialValue() { return new Context(); }
   };

   /** 
    * The original variable set to the parameter.  Is set by the c'tor or on the first call of 
    * {@link #get(int)}.
    */
   private T[] variable = null;
   
   /** Flag to identify if {@link #get(int)} is called more than once. */
   private boolean getCalled = false;

   /** The latest parameter reference. */
   private T[] parameter = null;
   
   /**
    * Constructor.
    */
   public OutputExtentParameter()
   {
      Context local = context.get();
      Scope scope = local.getCurrentScope();
      scope.addAssigner(this);
      
      // save the original array, in case it gets changed (at this time, the only way to change
      // references is by a resize).
      this.variable = getVariable();
   }
   
   /**
    * Get the variable reference, passed to the function/procedure's parameter. Emitted and to be 
    * used only in cases of OUTPUT and INPUT-OUTPUT extent parameters.
    * 
    * @return   See above.
    */
   public abstract T[] getVariable();

   /**
    * Change the caller's reference to the new reference.  Emitted and to be used only in cases of
    * OUTPUT and INPUT-OUTPUT extent parameters.
    * 
    * @param    reference
    *           The new reference.
    */
   public abstract void setVariable(T[] reference);

   /**
    * Special get which is emitted when the parameter is fixed-extent.  If the received array is
    * dynamic-extent and not yet initialized, then this will set the {@link #variable}'s extent 
    * using the parameter's extent.
    * 
    * @param    extent
    *           The parameter's extent set at the function definition.
    *           
    * @return   The reference, which might have been resized.
    */
   public T[] getVariable(int extent)
   {
      T[] ref = variable;

      if (getCalled)
      {
         return ref;
      }

      getCalled = true;

      // if the passed argument is a not-yet-initialized dynamic-extent array, we set its size
      // now
      if (ref.length == 0)
      {
         ref = ArrayAssigner.resize(ref, extent);
         
         // this does not update the variable too, because in case of error, the variable must not
         // get changed
      }
      
      
      return ref;
   }
   
   /**
    * Set the current reference for the parameter. As this is mutable, we need to save it each
    * time the parameter's reference gets changed.
    */
   public void setParameter(T[] parameter)
   {
      this.parameter = parameter;
   }

   /**
    * Assign the current value of the parameter variable back to the caller's reference.  This is
    * expected to occur upon exit from the function or procedure with which this object is
    * associated.
    */
   private void assign()
   {
      if (parameter == null)
      {
         // TODO: throw exception
         throw new RuntimeException("The extent parameter reference was not set," +
         		"                    before the function/procedure end !");
      }
      
      if (parameter != variable)
      {
         // if the reference has changed, update the with a copy of the latest data
         setVariable(ArrayAssigner.copyOf(parameter));
      }
   }
   
   /**
    * Object which manages the context local state of field assigners.  It
    * manages a deque of scopes, each of which contains zero or more
    * <code>FieldAssigner</code> objects.  As the currently executing
    * program's block scopes open and close, these scopes are pushed and
    * popped from the deque.
    */
   private static class Context
   implements Scopeable
   {
      /** Deque containing field assigners by scope */
      private final ArrayDeque<Scope> scopes = new ArrayDeque<Scope>();
      
      /**
       * Default constructor.
       */
      private Context()
      {
         scopes.push(new Scope());
      }
      
      /**
       * Called as each block scope opens, at which time we push a scope onto
       * the deque.
       * 
       * @see com.goldencode.p2j.util.Scopeable#scopeStart()
       */
      public void scopeStart()
      {
         scopes.push(new Scope());
      }
      
      /**
       * Called as each block scope closes, at which time we pop a scope from
       * the deque.
       * 
       * @see com.goldencode.p2j.util.Scopeable#scopeFinished()
       */
      public void scopeFinished()
      {
         scopes.pop();
      }
      
      /**
       * Get the object which manages field assigners for the current scope.
       * 
       * @return  Current scope.
       */
      private Scope getCurrentScope()
      {
         return scopes.peek();
      }
   }
   
   /**
    * One instance of this class is created for each block scope which the
    * currently executing program enters.  It contains zero or more field
    * assigners specific to that scope.
    * <p>
    * After exiting a function or procedure, the {@link #processAssignments()}
    * method is invoked, which causes each field assigner to assign the
    * current value of its parameter value back to the associated DMO
    * property.
    */
   private static class Scope
   implements OutputParameterAssigner
   {
      /** List of field assigners, created lazily */
      private List<OutputExtentParameter> list = null;
      
      /**
       * Discard any field assigners.
       */
      public void abort()
      {
         list = null;
      }
      
      /**
       * Cause each (if any) field assigner in the current scope to assign the
       * current value of its paramter value back to the DMO property with
       * which it is associated.
       */
      public void processAssignments()
      {
         if (list == null)
         {
            return;
         }

         boolean batch = (list.size() > 1);
         
         if (batch)
         {
            RecordBuffer.startBatch();
         }
         
         try
         {
            for (OutputExtentParameter next : list)
            {
               next.assign();
            }
         }
         finally
         {
            list = null;
            
            if (batch)
            {
               RecordBuffer.endBatch();
            }
         }
      }
      
      /**
       * Add a field assigner to this scope.
       * 
       * @param   assigner
       *          Field assigner.
       */
      private void addAssigner(OutputExtentParameter assigner)
      {
         if (list == null)
         {
            list = new ArrayList<OutputExtentParameter>(2);
            TransactionManager.registerOutputParameterAssigner(this);
         }
         
         list.add(assigner);
      }
   }
}
